
(* Runs passes of optimizations. - Tom 7 *)
structure Driver :> DRIVER =
struct

  val debug = 
     Params.flag false
       (SOME("-debug-driver", "turn on debugging for the optimization driver"))
       "debug-driver"

  val doprint =
     Params.flag false
       (SOME("-print-steps", "print after each optimization"))
       "print-steps"

  val onlydo =
     Params.param ""
       (SOME("-only", "comma separated list: only do these opts"))
       "onlydo"

  (* if fewer than this amount of time left (plus the time it 
     should take to do the next opt), give up. *)
  val minsecleft = 10

  fun ++ r = r := !r + 1

  fun dbprint s = if !debug then (TextIO.output (TextIO.stdErr, s); 
                                  TextIO.output (TextIO.stdErr, "\n")) else ()

  datatype optimization =
    Iltoil of Il.exp -> Il.exp
  | Liltolil of Lil.exp -> Lil.exp
  | Iltolil of Il.exp -> Lil.exp
  | Liltoil of Lil.exp -> Il.exp
(* ? perry's thing *)

  type opt = {optname : string, disabled : bool ref,
              trusted : bool, blowup : bool,
              timesrun : int ref,
              scorefn : int -> int, 
              speed : int ref,
              score : int ref,
              optfun : optimization,
              lastcall : int ref}

  val opts = (ref nil) : opt list ref

  fun register_opt {optname, disabled, trusted, 
                    blowup, optfun, initscore,
                    scorefn} =
    let 
    in
      opts := ({optname=optname, disabled=ref disabled, trusted=trusted,
                lastcall=ref (Option.getOpt (initscore, 9999999)), 
                scorefn=scorefn,
                blowup=blowup, optfun=optfun, timesrun=ref 0, 
                score=ref 0,
                speed=ref 1} : opt) :: !opts
    end

  datatype exp = IL of Il.exp | LIL of Lil.exp 

  fun measure (IL e) = Print.sizet e
    | measure (LIL e) = Print.sizel e

  fun disabler nil = ()
    | disabler (({optname,disabled,...}: opt) ::t) =
      if !onlydo = "" then ()
      else 
        let val l = (String.tokens (StringUtil.ischar #",") (!onlydo))
        in if not (List.exists (fn i => i = optname) l) then
          disabled := true
           else ();
             disabler t
        end


  fun optable () =
    let 
      fun warnbool true = "**YES**"
        | warnbool false = ""

      val tab = map (fn {optname,disabled=ref disabled,
                         trusted,blowup,optfun,
                         timesrun=ref timesrun,
                         lastcall,
                         score=ref score,
                         speed=ref speed, ... } : opt =>
                     [optname, warnbool disabled, Bool.toString trusted,
                      Bool.toString blowup, Int.toString timesrun,
                      Int.toString score,
                      Int.toString speed]) (!opts)
      val tab = ["optname", "disabled", "trusted", "blowup", "timesrun",
                 "score", "maxtime"] :: tab
        
    in StringUtil.table 80 tab
    end

  fun debughead (seconds, exp) =
    (!debug) andalso
    let in
      dbprint ("optimize called with " ^ 
                       Int.toString seconds ^ " seconds.");
      dbprint ("The expression is of size " ^ 
                       Int.toString (Print.sizel exp) ^ " bytes.");
      dbprint ("The following optimizations are loaded: \n");
      dbprint (optable ());
      true
    end

  fun debugtail (seconds, exp) =
    let in
      dbprint ("Finished with " ^ Int.toString seconds ^ " seconds left.");
      dbprint ("Final expression size: " ^ Int.toString (measure exp) ^
               " bytes.");
      dbprint ("Optimization stats: \n");
      dbprint (optable());
      true
    end

  fun chooseopt (remain, sec, e) =
    let
      fun fb bestsc bo l nil = Option.map (fn a => (a, l)) bo
        | fb bestsc bo l ((h : opt as {disabled=ref false, 
                                       lastcall=ref lc, ...})::t) =
        if lc > bestsc then
          fb lc (SOME h) (case bo of NONE => l | SOME p => p :: l) t
        else fb bestsc bo (h::l) t
        | fb bestsc bo l (h::t) = fb bestsc bo l t
    in
      fb (~1) NONE nil remain
    end

  fun gsec t = LargeInt.toInt (Time.toSeconds (Timer.checkRealTimer t))

  fun printmod exp =      
    case exp of 
        IL  e => Print.treeoutput e
      | LIL e => Print.linearoutput e

  (* XXX remove ! *)
  fun modequal (IL e1, IL e2) =
        TilValidate.equiv (e1, e2)
    | modequal (LIL e1, LIL e2) =
        LilValidate.equiv (e1, e2)
    | modequal (IL e1, LIL e2) =
        TilValidate.equiv (e1, ToTree.totree e2)
    | modequal (LIL e1, IL e2) =
        TilValidate.equiv (ToTree.totree e1, e2)

  fun samemeaning (mng, IL e) =
          TilValidate.verify_meaning mng e
    | samemeaning (mng, LIL e) =
          LilValidate.verify_meaning mng e


  fun optloop ((tried, remain), timer, seconds, mng, exp, expsize) =
    let 
      val stime = gsec timer
      val maxopttime = 
        foldl (fn ({speed=ref m, ...},b) => Int.max(m, b)) 1 (!opts)
    in
      if stime > (seconds - (minsecleft + maxopttime)) then 
        let in
          dbprint "Ran out of time!\n";
          !debug andalso debugtail (seconds - stime, exp);
          printmod exp
        end
      else
        case chooseopt (remain, seconds - stime, exp) of
          NONE => (* none left!! *)
            let in
              !debug andalso debugtail (seconds - stime, exp);
              printmod exp
            end
        | SOME (opt as {optfun=ff,optname,trusted, speed,
                        timesrun, score, lastcall, scorefn, 
                        disabled,...} : opt, remain) => 
            (* do a translation if necessary; call optimization. *)
            let 
              val _ = dbprint ("Running optimization '" ^ optname ^ "'... (" ^
                               Int.toString (seconds - stime) ^ 
                               " sec. remain)")
              val res = 
                (case exp of
                   IL exp => 
                     (case ff of 
                        Iltoil f => IL (f exp)
                      | Iltolil f => LIL (f exp)
                      | Liltoil f => IL (f (ToLinear.tolinear exp))
                      | Liltolil f => LIL (f (ToLinear.tolinear exp)))
                 | LIL exp => 
                        (case ff of
                           Iltoil f => IL (f (ToTree.totree exp))
                         | Iltolil f => LIL (f (ToTree.totree exp))
                         | Liltoil f => IL (f exp)
                         | Liltolil f => LIL (f exp))) 
                   handle e =>
                     let in
                       dbprint ("EXCEPTION: " ^ exnMessage e ^ "!");
                       (* return an empty module so it will be
                          flagged as failing *)
                       IL (Il.Seq nil)
                     end

              val _ = if !doprint then 
                        let in
                           printmod res;
                           print "\n";
                           print (Print.treestructure 
                                  (case res of IL e => e 
                                       | LIL e => ToTree.totree e));
                           print "\n"
                        end else ()


              val optelap = gsec timer - stime

              val ok = trusted orelse samemeaning (mng, res)
              val nowsize = measure res

              (* heavyweight work is done, now measure elapsed time *)
              val elapsed = gsec timer - stime

              val loss = expsize - nowsize
            in
              speed := Int.max (!speed, elapsed);
              lastcall := scorefn loss;
              ++ timesrun;
              if ok then
                let in
                  dbprint ("Elapsed: " ^ Int.toString elapsed ^ " sec. [" ^
                           Int.toString (elapsed - optelap) ^ 
                           " to verify/measure]");
                  score := !score + loss;
                  if loss > 0 then
                    let in
                      dbprint ("Reduced size by " ^ Int.toString loss ^ 
                               " bytes: " ^ Int.toString nowsize ^ "");
                      optloop ((nil, remain @ tried @ [opt]), 
                               timer, seconds, mng, res, nowsize)
                    end
                  else (* XXX use old exp, or new here? *) 
                    let in
                      dbprint ("Wasn't able to reduce: " ^ 
                               Int.toString (~loss) ^
                               " byte increase!");
                      optloop ((tried @ [opt], remain), 
                               timer, seconds, mng, res, nowsize)
                    end
                end
              else
                let in
                  (* optimization FAILED! *)
                  (* XXX perhaps we don't always want to throw it away?
                     like, if it has a high score already? *)
                  dbprint ("********* OPTIMIZATION FAILURE: " ^ 
                           optname ^ " !! *********");
                  disabled := true;
                  optloop ((tried, remain), timer, seconds, mng,
                           exp, expsize)
                end
            end
    end

  fun optimize {timer, seconds, exp} =
    let 
      val _ = debughead (seconds, exp)
      val _ = disabler (!opts)
      val mng = LilValidate.convert (Meaning.null, exp)
    in
      optloop ((nil, !opts), timer, seconds, 
               mng,
               LIL exp, Print.sizel exp)
    end

  fun drivers () = map (fn {optname, optfun, ...} : opt => 
                          (optname, optfun)) (!opts)

end
