
structure FindFail =
struct

  local
  open Token Il
  in
    exception Done
  val debug = 
     Params.flag false
       (SOME("-debug", "turn on debugging (print steps)"))
       "debug"

  val ast = 
     Params.flag false
       (SOME("-ast", "output the ast instead of the text"))
       "ast"

  val sbet =
    Params.param "20"
       (SOME("-sbetween", "every n seconds, print out status (expensive)"))
       "sbet"

  fun eprint s = (TextIO.output (TextIO.stdErr, s))
  fun deprint s = if !debug then eprint s else ()
  fun detree red = deprint (Print.treetostring red);

  fun ok f exp =
    let val res = f exp
    in TilValidate.equiv (res, exp)
    end handle _ => false

  local val v : Word32.word ref = 
              ref (Word32.xorb
                   (Posix.Process.pidToWord (Posix.ProcEnv.getpid ()),
                    Word32.fromInt
                    (LargeInt.toInt (Time.toSeconds (Time.now ())))))
  in
    fun randn () = 
      let 
        val vv = !v
        val x = (vv * 0w28574) + 0w39852
        val x = Word32.xorb (x, 0w2309761)
        val x = x mod 0w65537
      in
        v := x;
        abs (Word32.toInt (Word32.andb(x, 0w65535)))
      end
  end

  fun dropnth 0 (h::t) = t
    | dropnth _ nil = nil
    | dropnth n (h::t) = h :: (dropnth (n - 1) t)

  fun maxdepth (Text _) = 1
    | maxdepth (Seq l) = 1 + foldl Int.max 0 (map maxdepth l)
    | maxdepth (Tag (_, e)) = 1 + maxdepth e

  fun overtwo 1 = 1
    | overtwo n = n div 2

  fun mapmaybe f l =
    let 
      val tomap = (randn () mod (overtwo (length l))) + 1
      fun mm _ nil = nil
        | mm 0 (h::t) = t
        | mm n (h::t) = 
        if randn () mod 100 <= 10 then f h :: mm (n-1) t
        else h :: mm n t
    in
      mm tomap l
    end

  (* chop "randomly" prunes the tree. *)
  fun chop _ (Text "") = Seq nil
    | chop _ (Text s) =
    (Text (* make sure we make it smaller *)
     (case (randn () mod size s) mod 5 of
        0 => if randn () mod 100 < 10 then "" else "*"
      | 1 => "*"
      | 2 => " *"
      | 3 => "* "
      | _ => " * "))
    | chop n (Seq (Seq l :: r)) = chop n (Seq (l @ r))
    | chop _ (Seq nil) = Seq nil
    | chop 0 (Seq el) =
    (* drop some arms of the sequence randomly. *)
      let 
        val todrop = (randn () mod (overtwo (length el))) + 1
        fun dp 0 l = l
          | dp n l =
          let val i = randn () mod length l
          in dp (n - 1) (dropnth i l)
          end
      in Seq (dp todrop el)
      end
    | chop n (Seq l) = Seq (mapmaybe (chop (n-1)) l)
    | chop n (Tag (nil, e)) = chop n e
    | chop 0 (Tag (tl, e)) =
      (* drop some tags randomly. *)
      let 
        val todrop = (randn () mod (overtwo (length tl))) + 1
        fun dp 0 l = l
          | dp n l =
          let val i = randn () mod length l
          in dp (n - 1) (dropnth i l)
          end
      in Tag (dp todrop tl, e)
      end
    | chop n (Tag (tl, e)) =
      Tag (tl, chop (n - 1) e)

  fun find f exx =
    let
      val md = maxdepth exx
      val _ = eprint ("Depth: " ^ Int.toString md ^ "\n")

      val sbetween = Option.getOpt (Int.fromString(!sbet), 20)

      val tmr = Timer.startRealTimer ()
      fun gsec () = LargeInt.toInt (Time.toSeconds (Timer.checkRealTimer tmr))

      (* if we've gone 25 tries without finding a valid chop,
         increase depth. *)
      fun ff tot d 25 exp norep oldsize = 
                      let in
                        eprint ("Proceeding to depth " ^ 
                                Int.toString (d + 1) ^ ". (now at size " ^
                                Int.toString (oldsize) ^ ")\n");
                        ff tot (d + 1) 0 exp false oldsize
                      end
        | ff tot d tries exp norep oldsize =
            if norep andalso (true orelse (randn () mod 1000) < 10) then
              let val s = gsec ()
              in
                if s > (tot + sbetween) then
                  let in
                    eprint ("[" ^ Int.toString s ^ " sec. Depth " ^
                            Int.toString d ^ " tries: " ^ 
                            Int.toString tries ^ " size: " ^
                            Int.toString oldsize ^ "]\n");
                    ff s d tries exp false oldsize
                  end
                else ff tot d tries exp false oldsize
              end
            else
              let
        in
          if d > md then exp
          else 
            let
              val red = chop d exp
            in
              deprint ("Chopped expression: \n");
              detree red;
              deprint "\n";
              if ok f red then
                let in
                  deprint "BACKTRACK.\n";
                  (* oops, fixed it! *)
                  ff tot d (tries + 1) exp true oldsize
                end
              else 
                let val newsize = Print.sizet red
                in
                (* good, reduction with bug still *)
                if oldsize > newsize then
                  ff tot d 0 red true newsize
                else ff tot d (tries + 1) exp true oldsize
                end
            end
        end
      val smallest = ff (gsec ()) 0 0 exx false (Print.sizet exx)
    in
      smallest
    end


  fun findfail (f, exx) =
    if ok f exx then
      let in
        eprint "But that optimization is ok on that exp!\n";
        raise Done
      end
    else 
      let in
        eprint ("(Random number: "  ^ Int.toString (randn ()) ^ ")\n");
        find f exx
      end

  fun run () =
    case Params.docommandline () of
      [optname] =>
        (case List.find (fn (oname,Driver.Iltoil ofun) => oname = optname 
                          | _ => false) 
                    (Driver.drivers ()) of
          SOME (on,Driver.Iltoil f) => 
        let
          val exx = ToTree.totree ((Tokenize.tokenize (FStream.stdin ())))
          val sm = findfail (f, exx)
        in
           if !ast then
             let in
               print "local open Token Il in\nval ast = \n";
               print (Print.treestructure sm);
               print "\nend;\n"
             end
           else Print.treeoutput sm
        end
         | _ => eprint 
         "No optimization with that name was found (it must be il -> il).\n")
    | _ => 
           let in
             eprint ("try: findfail \"optname\" < testfile\n");
             eprint (Params.usage ());
             eprint "\n"
           end
  end

end

val _ = FindFail.run ()
